#ifndef __A4C_IME_H
#define __A4C_IME_H

#include "iome.h"
#include <stdarg.h>
#include <ctype.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <X11/Xlibint.h>
#include <X11/xpm.h>
#include <dlfcn.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  unsigned long foreground;
  unsigned long background;
  unsigned long lightcolor;
  unsigned long darkcolor;
  unsigned long englishcolor;
  char *        fgcolor;
  char *        ltcolor;
  char *        dkcolor;
  char *        bgcolor;
  char *        egcolor;
} a4cEnvColor;

typedef struct {
  int x1;
  int y1;
  int x2;
  int y2;
} a4cRECT;

#define A4C_STYLE_NOBORDER 1
#define A4C_STYLE_CENTER   2
#define COMPARE_FLAG(a,b) ((a&b)==b)

typedef struct {
  a4cRECT Border;
  a4cRECT Text;
  int     style;
  char *  message;
} a4cKeyBox;

int isInRect(a4cRECT *RECT,int x,int y);
int isInKeyBox(a4cKeyBox *KEYBOX,int x,int y);

/**************************************************\
                        Utils
\**************************************************/
void SwapEndian(char *i);
void oops(char * fmt,...);
void nope(char * fmt,...);
void die(char * fmt,...);

#ifdef NO_MESSAGE_MODE
# define report nope
#else
# define report oops
#endif

#define IS_ENDIAN()   (*(int*)"\x11\x22\x33\x44"==0x11223344)

char * xDisplay(void);

int xChangeTitle(Display * disp, Window win, char *title);
int a4cPlayWAVE(char *DEVICE,char *FILENAME);

/*******************************************************\
                      MAIN Functions

\*******************************************************/

typedef struct {
  int state;
  int key;
  int method;
  char * command;
} Accelerator;

#define A4C_XIN_ATOM  "XCIN_A"
/* This structure no long used in the A4C IME */
typedef struct {
  char kb_state;
  char _CurInMethod;
} InmdState;

#define XCIN_KEY_LEN 512
typedef struct {
  int       len;
  int       status;
  InmdState inmdstate;
  char      key[XCIN_KEY_LEN];
} XCIN_RES;

typedef char * a4cSTRING;

typedef void * a4cHANDLE;

typedef struct {
  int           Endian;
  Display *     disp;
  Window        root;
  Window        win;
  Window        CLI_WIN;
  Window        LastWIN;
  Colormap      COLORMAP;
  int           screen;
  char *        title;

  GC            EGC;
  GC            CGC;
  XFontStruct * EFONT;
  XFontStruct * CFONT;

  Atom          XCIN;
  Atom          WM_DEL_WINDOW;

  a4cEnvColor   ENVCOLOR;

  a4cRECT       Border;
  a4cKeyBox     Method;
  a4cKeyBox     Charset;
  a4cKeyBox     FontSize;
  a4cKeyBox     KeyArea;
  a4cKeyBox     ListArea;
  iomeEncoding  CC;

  XWindowAttributes attr;
  XClassHint    CLASS;
  XWMHints      wm;

  Pixmap        ICON;
  XSizeHints    SIZE;

  int           Hide;
  int           fontsize;
  a4cSTRING     fontsize_name[2];

  char *        base_charset;

  int           method_size;
  int           current_method;
  int           last_method;
  iomeModule *  METHOD_NOW;
  iomeModule ** METHOD;
  char *        METHOD_ENGLISH;

  int           filter_size;
  int           current_filter;
  int           last_filter;
  int           translating;
  iomeModule *  FILTER_NOW;
  iomeModule ** FILTER;
  char *        FILTER_NEUTRAL;

  char *        current_charset;
  char *        neutral_charset;

  int           dl_handles;
  a4cHANDLE *   SO_HANDLES;  

  XCIN_RES      xcin_res;
  int           replied;

  /* Sound Settings */
  char *        AUDIO_DEVICE;
  char *        WARNING_FILE;
  
  /* AutoRaise - Hide */
  int           autoraise;
  int           autohide;
  int           autostate;
  
  /* AcceleratorKey */
  int           AccelKeys;
  Accelerator   ACCEL[512];
} a4cMAIN;

#define A4C_METHOD_WIDTH   5
#define A4C_CHARSET_WIDTH  3
#define A4C_FONTSIZE_WIDTH 2
#define A4C_KEYAREA_WIDTH  12
#define A4C_MARGAIN        2

#define A4C_EVENTS             \
        KeyPressMask         | \
        FocusChangeMask      | \
        StructureNotifyMask  | \
        ButtonPressMask      | \
        ExposureMask         | \
        PropertyChangeMask   | \
        VisibilityChangeMask

void a4cErrorHandler(Display *disp,XErrorEvent *event);
int a4cWarmUp(
               int argc,
               char **argv,
               a4cMAIN *MAIN,
               iomeDefaultConf *CONF,
               char **pixmap
             );

typedef struct {
  a4cMAIN MAIN;
  iomeDefaultConf CONF;
} a4cALL;

typedef void (*a4cEVENT)(XEvent event,a4cALL *ALL);

typedef struct {
  a4cEVENT keypress;
  a4cEVENT clientmessage;
  a4cEVENT propertynotify;
  a4cEVENT mappingnotify;
  a4cEVENT graphicsexpose;
  a4cEVENT expose;
  a4cEVENT buttonpress;
  a4cEVENT visibilitynotify;
  a4cEVENT destroynotify;
  a4cEVENT unknown;
} a4cEventHandler;

void a4cXWAIT(a4cEventHandler *a4c,a4cALL *ALL);
/* Default Functions */
void a4cKeyPress(XEvent event,a4cALL *ALL);
void a4cMappingNotify(XEvent event,a4cALL *ALL);
void a4cUnknownEvent(XEvent event,a4cALL *ALL);

void a4cDrawShadowBox(a4cMAIN *MAIN,a4cRECT * RECT, int up_down);
void a4cDrawTextInBox(a4cMAIN *MAIN,a4cKeyBox * KeyBox);
void a4cDrawTextInCenterBox(a4cMAIN *MAIN,a4cKeyBox * KeyBox);
void a4cDraw(a4cMAIN *MAIN);

void a4cMakeNoise(a4cMAIN *MAIN);

#define A4C_NOWHERE  0
#define A4C_METHOD   1
#define A4C_CHARSET  2
#define A4C_FONTSIZE 3
#define A4C_KEYAREA  4
#define A4C_LISTAREA 5
int  a4cWhere(a4cMAIN *MAIN,int x,int y);
void a4cLoadAccelerator(char *filename,a4cMAIN *MAIN);
int  a4cLookupAccelerator(int state,int ch,a4cMAIN *MAIN);

/***************************************\
          XCIN Compatible
\***************************************/
#define A4C_REJECT    (0)
#define A4C_ACCEPT    (1)
#define CLI_FOCUS_IN  (1)
#define LOAD_KEYWORD  (2)
#define CLI_FOCUS_OUT (3)

void WriteReply(a4cMAIN *MAIN,char *str,int len,int status);
void SendChineseKey(a4cMAIN *MAIN,unsigned char *s);
void SendString(a4cMAIN *MAIN,unsigned char *s);
void SendChar(a4cMAIN *MAIN,char ch);
void SendZeroOut(a4cMAIN *MAIN);
void SendReject(a4cMAIN *MAIN);

/**************************************\
              Load Modules
\**************************************/
int a4cLoadModules(a4cMAIN *MAIN,iomeConfig *CONF,iomeConfig *CFG);


/**************************************\
          Charset Toolkit
\**************************************/
int UnifyCharset(a4cMAIN *MAIN,char *str);


#ifdef __cplusplus
}
#endif

#endif
