#include "iome.h"

typedef struct {
  int        Size;
  int        Side;
  char       KEY[12];  /* Five is enough */
  iomeString CODE[16]; /* Full Font for 0-f */
} IntCode;

int imeRegister(iomeSysInfo *INFO,iomeRegistry *REG)
{
  char *Accept_System[]={
    "x",
    "tty",
    NULL
  };
  char **IME = Accept_System;
  int n = 0;
  while (n>=0&&IME[n]) {
    if (strcasecmp(INFO->system,IME[n])==0) n=-2;
    n++;
  }
  if (n>=0) return 0;
  REG->encoding      = "big5";
  REG->class         = "ime";
  REG->accept_system = Accept_System;
  return 1;
}

int imeInit(iomeConfig *CONF,iomeRecord *RECORD)
{
  IntCode * IC;
  iomeString HEX[]={
    "0","1","2","3","4","5","6","7",
    "8","9","A","B","C","D","E","F"
  };
  int n;
  char * HEXFILE;
  iomeConfig CONFIG;

  IC = (IntCode *)malloc(sizeof(IntCode));
  RECORD->MethodName = iomeWhichItem(CONF,"method.intcode");
  RECORD->FilterName = NULL;
  RECORD->Encoding   = "big5";
  RECORD->ListArea   = (char *)malloc(64); RECORD->ListArea[0]=0;
  RECORD->KeyArea    = (char *)malloc(64); RECORD->KeyArea[0]=0;
  RECORD->Value      = (char *)malloc(12); RECORD->Value[0]=0;
  RECORD->Owner      = IC;

  HEXFILE = iomeWhichItem(CONF,"hexcode.big5");
  iomeLoadConfiguration(HEXFILE,&CONFIG);
  IC->Size = 0;
  IC->Side = 0;
  for (n=0;n<16;n++) 
    IC->CODE[n]=strdup(iomeWhichItem(&CONFIG,HEX[n]));
  iomeFreeConfiguration(&CONFIG);
  return 1;
}

int imePrepare(iomeRecord *RECORD)
{
  IntCode *IC = RECORD->Owner;
  IC->Size = 0;
  IC->Side = 0;
  RECORD->ListArea[0]=0;
  RECORD->KeyArea[0]=0;
  RECORD->Value[0]=0;
  return 1;
}

void MakeID(IntCode *IC)
{
  switch(IC->Size) {
    case 1: IC->KEY[1]=1;
    case 2: IC->KEY[2]=4;
    case 3: IC->KEY[3]=IC->Side ? 8:0; break;
  }
}

void MakeIndex(iomeRecord *RECORD)
{
  char HEX[]="0123456789ABCDEF";
  IntCode *IC = RECORD->Owner;
  int c,x,i;

  if (IC->Size==4||IC->Size==0) return;

  c = ((int)IC->KEY[0])*16+((int)IC->KEY[1]);
  x = ((int)IC->KEY[2])*16+((int)IC->KEY[3]);
  i = IC->KEY[3];

  sprintf(RECORD->ListArea,
    "%c%c%cx :  %c%c%c %c%c%c %c%c%c %c%c%c %c%c%c %c%c%c %c%c%c %c%c%c   <:>",
    HEX[IC->KEY[0]],HEX[IC->KEY[1]],HEX[IC->KEY[2]],
    HEX[i  ],c,x,
    HEX[i+1],c,x+1,
    HEX[i+2],c,x+2,
    HEX[i+3],c,x+3,
    HEX[i+4],c,x+4,
    HEX[i+5],c,x+5,
    HEX[i+6],c,x+6,
    HEX[i+7],c,x+7
  );
}

void MoveToNext(IntCode *IC)
{
  if (IC->Side==0) { IC->Side=1; IC->KEY[3]=8; return; }
  IC->Side = 0; IC->KEY[3]=0;
  if (IC->Size==3) return;
  if (IC->KEY[2]==7 ) { IC->KEY[2]=10; return; }
  if (IC->KEY[2]==15) { IC->KEY[2]=4 ; return; }
  IC->KEY[2]++;
}

void MoveToPrevious(IntCode *IC)
{
  if (IC->Side==1) { IC->Side=0; IC->KEY[3]=0; return; }
  IC->Side = 1; IC->KEY[3]=8;
  if (IC->Size==3) return;
  if (IC->KEY[2]==10) { IC->KEY[2]=7 ; return; }
  if (IC->KEY[2]==4 ) { IC->KEY[2]=15; return; }
  IC->KEY[2]--;
}

int imeFeedKey(iomeRecord *RECORD,int ch,int state)
{
  IntCode *IC = (IntCode*)RECORD->Owner;
  int key = iomeTOUPPER(ch);
  if (IC->Size==0) {
    RECORD->KeyArea[0]=0;
    RECORD->ListArea[0]=0;
    IC->Side = 0;
  }
  if (key==IME_KEY_BACK||key==IME_KEY_DELETE) {
    /* Delete A Key */
    if (IC->Size==0) return IME_RT_REJECT;
    IC->Size--;
    RECORD->KeyArea[IC->Size*2]=0;
    return IME_RT_BLOCKOUT;
  }
  if (key=='>'&&IC->Size>0) {
    MoveToNext(IC);
    MakeIndex(RECORD);
    return IME_RT_BLOCKOUT;
  }
  if (key=='<'&&IC->Size>0) {
    MoveToPrevious(IC);
    MakeIndex(RECORD);
    return IME_RT_BLOCKOUT;
  }
  if (!((key>='0'&&key<='9')||(key>='A'&&key<='F'))) return IME_RT_REJECT;
  if (key>='A'&&key<='F') key=key-'A'+10; else key -='0';
  if (IC->Size==0&&key<10)                   return IME_RT_REJECT;
  if (IC->Size==1&&IC->KEY[0]==15&&key==15)  return IME_RT_BLOCKOUT;
  if (IC->Size==2&&(key<4||(key>7&&key<10))) return IME_RT_BLOCKOUT;
  if (IC->Size<4) IC->KEY[IC->Size]=key;
  strcat(RECORD->KeyArea,IC->CODE[key]);
  IC->Size++;
  /* Make Index */
  MakeID(IC);
  MakeIndex(RECORD);
  if (IC->Size==4) {
    RECORD->Value[0]=(char)(((int)IC->KEY[0])*16+(int)IC->KEY[1]);
    RECORD->Value[1]=(char)(((int)IC->KEY[2])*16+(int)IC->KEY[3]);
    RECORD->Value[2]=0;
    strcpy(RECORD->ListArea,RECORD->Value);
    IC->Size = 0;
    return IME_RT_SEND;
  }
  return IME_RT_BLOCKOUT;
}

int imeListArea(iomeRecord *RECORD,int x,int y)
{
  return 0;
}

int imeKeyArea(iomeRecord *RECORD,int x,int y)
{
  return 0;
}

int imeFocus(iomeRecord *RECORD)
{
  return 0;
}

int imeQuit(iomeRecord *RECORD)
{
  IntCode *IC = RECORD->Owner;
  int n;
  for (n=0;n<16;n++) free(IC->CODE[n]);
  free(IC);
  free(RECORD->ListArea);
  free(RECORD->KeyArea);
  free(RECORD->Value);
  return 1;  
}

