#include "iome.h"

#define PHONE_FIRST  21
#define PHONE_SECOND 24
#define PHONE_THIRD  37
#define PHONE_SOUND  41

/*
  Hash Table
  22 4 14 5 = 6160 * 4 = 24KB
  KEY = ((KEY[0]*4+KEY[1])*14+KEY[2])*5+KEY[3]
*/

#define HASH_SIZE 6160

int SK(int c)
{
  if (c==0) return 0;
  if (c<=PHONE_FIRST)  return c;
  if (c<=PHONE_SECOND) return c-PHONE_FIRST;
  if (c<=PHONE_THIRD)  return c-PHONE_SECOND;
  return c-PHONE_THIRD;
}

int GetHashKey(char *BUF)
{
  return (((((SK(BUF[0])*4)+SK(BUF[1]))*14)+SK(BUF[2]))*5)+SK(BUF[3]);
}

char * PhoneTAB=" 1qaz2wsxedcrfv5tgbyhnujm8ik,9ol.0p;/-7634";
typedef struct {
  char PhoneKEY[5];
  char CODE[3];
} PhoneMAP;

typedef struct {
  int        Size;
  int        Select;
  int        Cleanup;
  char       KEY[12];
  char       INDEX[60];
  iomeString CODE[60];
  iomeString SPACE;
  int        SIZE;
  PhoneMAP * MAP;
  int        HashSize;
  int *      RangeTable; /* Record */
  int        HASH[HASH_SIZE];
  int        from;
  int        now;
  int        to;
} PhoneticMethod;

int imeRegister(iomeSysInfo *INFO,iomeRegistry *REG)
{
  char *Accept_System[]={
    "x",
    "tty",
    NULL
  };

  char **IME = Accept_System;
  int n = 0;
  while (n>=0&&IME[n]) {
    if (strcasecmp(INFO->system,IME[n])==0) n=-2;
    n++;
  }
  if (n>=0) return 0;
  REG->encoding      = "big5";
  REG->class         = "ime";
  REG->accept_system = Accept_System;
  return 1;
}

int FindPhoneKEY(int c)
{
  int n = 0;
  while (PhoneTAB[n]!=0) {
    if (PhoneTAB[n]==c) return n;
    n++;
  }
  printf("Phony!\n");
  return -1;
}

void ConvertPhoneKEY(char *BUF)
{
  BUF[0] = (char)FindPhoneKEY(BUF[0]);
  BUF[1] = (char)FindPhoneKEY(BUF[1]);
  BUF[2] = (char)FindPhoneKEY(BUF[2]);
  BUF[3] = (char)FindPhoneKEY(BUF[3]);
}

int GetLine(FILE *fp,char *BUF)
{
  int c;
  while ((c=fgetc(fp))!=EOF&&c!=10) if (c!=13) *BUF++ = c;
  *BUF = 0;
  return (c!=EOF);
}

int imeInit(iomeConfig *CONF,iomeRecord *RECORD)
{
  PhoneticMethod * PM;
  char BUF[64];
  char HAHA[5];
  char *TAB;
  FILE *fp;
  int n,c;

  PM = (PhoneticMethod *)malloc(sizeof(PhoneticMethod));
  memset(PM,0,sizeof(PhoneticMethod));

  RECORD->MethodName = iomeWhichItem(CONF,"method.phonetic");
  RECORD->FilterName = NULL;
  RECORD->Encoding   = "big5";
  RECORD->ListArea   = (char *)malloc(64); RECORD->ListArea[0]=0;
  RECORD->KeyArea    = (char *)malloc(64); RECORD->KeyArea[0]=0;
  RECORD->Value      = (char *)malloc(12); RECORD->Value[0]=0;
  RECORD->Owner      = PM;

  PM->Size   = 0;
  PM->Select = 0;
  /* Load Table */
  PM->SPACE = iomeWhichItem(CONF,"space");
  TAB = iomeWhichItem(CONF,"phonetic.kbm");
  if (TAB==NULL) {
    printf("Phonetic: Can not find kbm file\n");
    return 0;
  }
  fp = fopen(TAB,"rt");
  if (fp==NULL) {
    printf("Phonetic: Can not open kbm file %s\n",TAB);
  }
  n = 0;
  fscanf(fp,"%s",BUF); /* Read 0123456789 */
  while (fscanf(fp,"%s %c",BUF,&c)==2) {
    PM->INDEX[n] = c;
    PM->CODE[n]  = strdup(BUF);
    n++;
  }
  PM->INDEX[n]=0;
  fclose(fp);
  TAB = iomeWhichItem(CONF,"phonetic.table");
  if (TAB==NULL) {
    printf("Phonetic: Can not find table file\n");
    return 0;
  }
  fp = fopen(TAB,"rt");
  if (fp==NULL) {
    printf("Phonetic: Can not open table file %s\n",TAB);
    return 0;
  }
  /* Load Phonetic Table */
  PM->SIZE       = 0;
  PM->MAP        = NULL;
  PM->HashSize   = 0;
  PM->RangeTable = NULL;
  for (n=0;n<HASH_SIZE;n++) PM->HASH[n]=-1;
  memset(HAHA,0,5);
  while (GetLine(fp,BUF)) {
    PM->MAP = (PhoneMAP *)realloc(PM->MAP,sizeof(PhoneMAP)*(PM->SIZE+1));
    memcpy(&PM->MAP[PM->SIZE],BUF,8);
    PM->MAP[PM->SIZE].PhoneKEY[4] = 0;
    PM->MAP[PM->SIZE].CODE[2]     = 0;
    ConvertPhoneKEY(PM->MAP[PM->SIZE].PhoneKEY);
    /* Initialize Hash Table */
    if (memcmp(HAHA,PM->MAP[PM->SIZE].PhoneKEY,4)!=0) {
      memcpy(HAHA,PM->MAP[PM->SIZE].PhoneKEY,4);
      /* Add New Hash Entry */
      PM->RangeTable = (int*)realloc(PM->RangeTable,sizeof(int)*(PM->HashSize+1));
      PM->RangeTable[PM->HashSize]=PM->SIZE;
      PM->HASH[GetHashKey(HAHA)]=PM->HashSize;
      PM->HashSize++;
    }
    PM->SIZE++;
  }
  PM->RangeTable = (int*)realloc(PM->RangeTable,sizeof(int)*(PM->HashSize+1));
  PM->RangeTable[PM->HashSize] = PM->SIZE;
  PM->HashSize++;
  fclose(fp);
  return 1;
}

int imePrepare(iomeRecord *RECORD)
{
  PhoneticMethod *PM = RECORD->Owner;
  PM->Size    = 0;
  PM->Select  = 0;
  PM->Cleanup = 0;
  PM->from    = -1;
  PM->to      = -1;
  PM->now     = -1;
  memset(PM->KEY,0,4);
  RECORD->ListArea[0]=0;
  RECORD->Value[0]=0;
  sprintf(RECORD->KeyArea,"%s%s%s%s",PM->SPACE,PM->SPACE,PM->SPACE,PM->SPACE);
  return 1;
}

int ScanKBM(PhoneticMethod *PM,int key)
{
  int n=0;
  char *p = PM->INDEX;
  while (*p!=0&&*p!=key) { p++; n++; }
  if (PM->INDEX[n]==0) return -1;
  return n;
}

int FigureOutKey(iomeRecord *RECORD)
{
  PhoneticMethod *PM = (PhoneticMethod*)RECORD->Owner;
  int n = 0;
  int i,w;
  int from,to,stop;
  char BUF[64];

  i = PM->HASH[GetHashKey(PM->KEY)];
  if (i<0) {
    sprintf(RECORD->ListArea,"");
    PM->Select = 0;
    PM->from   = -1;
    PM->now    = -1;
    PM->to     = -1;
    return 0;
  }
  from = PM->RangeTable[i];
  to   = PM->RangeTable[i+1];
  n    = to - from;
  stop = (n<10) ? to:from+10;

  RECORD->ListArea[0]=0;
  for (i=from,w=1;i<stop;i++,w++) {
    strcpy(BUF,RECORD->ListArea);
    sprintf(RECORD->ListArea,"%s %s",BUF,PM->MAP[i].CODE);
  }
  strcpy(BUF,RECORD->ListArea);
  if ((to-from)>10) sprintf(RECORD->ListArea," :> %s",BUF); 
               else sprintf(RECORD->ListArea," :  %s",BUF);

  PM->from = from;
  PM->now  = from;
  PM->to   = to;

  return 1;
}

int IndexKey(iomeRecord *RECORD,int action)
{
  PhoneticMethod *PM = (PhoneticMethod*)RECORD->Owner;
  int n = 0;
  int i,w;
  int from,to,stop;
  char BUF[64];
  int p1,p2,p3;

  i = PM->HASH[GetHashKey(PM->KEY)];
  if (i<0) {
    sprintf(RECORD->ListArea,"");
    PM->Select = 0;
    PM->from   = -1;
    PM->now    = -1;
    PM->to     = -1;
    return 0;
  }
  from = PM->now;
  to   = PM->to;
  n    = to - from;
  switch (action) {
    case  0:
      from = PM->RangeTable[i];
      to   = PM->RangeTable[i+1];
      n    = to - from;
      PM->from = from;
      PM->to   = to;
      PM->now  = from;
    break;
    case -1: 
      if (PM->now>PM->from) {
        PM->now-=10;
        from = PM->now;
        n = to - from;
      }
    break;
    case  2: 
      if (PM->now+10>=PM->to) {
        PM->now = PM->from;
        from = PM->from;
        n = to - from;
        break;
      }
    case  1: 
      if (n>10) {
        PM->now+=10;
        from = PM->now;
        n = to - from;
      }
    break;
  }
  stop = (n<10) ? to:from+10;

  RECORD->ListArea[0]=0;
  for (i=from,w=1;i<stop;i++,w++) {
    strcpy(BUF,RECORD->ListArea);
    sprintf(RECORD->ListArea,"%s %d%s",BUF,w%10,PM->MAP[i].CODE);
  }

  strcpy(BUF,RECORD->ListArea);

  p1 = (PM->from!=PM->now)    ? '<':' ';
  p2 = ((PM->to-PM->from)>10) ? ':':' ';
  p3 = (PM->now+10<PM->to)    ? '>':' ';
  sprintf(RECORD->ListArea,"%c%c%c %s",p1,p2,p3,BUF);

  return 1;
}

int SendME(iomeRecord *RECORD,int id)
{
  PhoneticMethod *PM = (PhoneticMethod*)RECORD->Owner;

  if (id>=PM->SIZE) return IME_RT_BLOCKOUT;

  strcpy(RECORD->Value,PM->MAP[id].CODE);

  PM->Size    = 0;
  PM->Select  = 0;
  PM->Cleanup = 1;
  PM->now     = -1;
  PM->from    = -1;
  PM->to      = -1;
  strcpy(RECORD->ListArea,RECORD->Value);
  memset(PM->KEY,0,5);

  return IME_RT_SEND;
}

int imeFeedKey(iomeRecord *RECORD,int ch,int state)
{
  PhoneticMethod *PM = (PhoneticMethod*)RECORD->Owner;
  int key = iomeTOLOWER(ch);
  int id,n;

  if (PM->Cleanup) {
    sprintf(RECORD->KeyArea,"%s%s%s%s",PM->SPACE,PM->SPACE,PM->SPACE,PM->SPACE);
    PM->Cleanup =0;
  }

  switch (key) {
    case IME_KEY_ESC: /* Clean up the mess */
      if (PM->KEY[0]==0&&PM->KEY[1]==0&&
          PM->KEY[2]==0&&PM->KEY[3]==0) break;
      imePrepare(RECORD);
    return IME_RT_BLOCKOUT;
    case IME_KEY_BACK:
    case IME_KEY_DELETE:
      n=3;
      while (n>=0&&PM->KEY[n]==0) n--;
      if (n>=0&&PM->KEY[n]>0) {
        PM->KEY[n]=0;
        memcpy(&RECORD->KeyArea[n*2],PM->SPACE,2);
      } else break;
      FigureOutKey(RECORD);
      PM->Select = 0;
    return IME_RT_BLOCKOUT;
    case ' ': /* Okay */
      if (PM->now<0) break;
      /* Select Mode */
      if (PM->Select) IndexKey(RECORD,2);
                 else IndexKey(RECORD,0);
      PM->Select = 1;
      if (PM->from>0&&(PM->to-PM->from)==1) return SendME(RECORD,PM->from);
    return IME_RT_BLOCKOUT;
    case '<':
      if (PM->now<0) break;
      if (PM->now==PM->from) break;
      /* Select Mode */
      PM->Select = 1;
      IndexKey(RECORD,-1);
    return IME_RT_BLOCKOUT;
    case '>':
      if (PM->now<0) break;
      if ((PM->to-PM->now)<=10) break;
      /* Select Mode */
      PM->Select = 1;
      IndexKey(RECORD,1);
    return IME_RT_BLOCKOUT;
    default:
      if (ch>=0x100) return IME_RT_REJECT;
      if (PM->Select&&key>='0'&&key<='9'&&PM->now>0) {
        id = key -'0';
        id = (id==0) ? 9:id-1;
        id += PM->now;
        if (id<PM->to) return SendME(RECORD,id);
        return IME_RT_BLOCKOUT;
      } else {
        id = ScanKBM(PM,key);
        if (id<0) break;
        /* hit jackpot */
        if (id<PHONE_FIRST)  n=0; else
        if (id<PHONE_SECOND) n=1; else
        if (id<PHONE_THIRD)  n=2; else
        if (id<PHONE_SOUND)  n=3;
        memcpy(&RECORD->KeyArea[n*2],PM->CODE[id],2);
        PM->KEY[n]=id+1;
        if (n==3) {
          /* Entering Selection Mode */
          if (IndexKey(RECORD,0)) PM->Select = 1;
          if (PM->from>0&&(PM->to-PM->from)==1) return SendME(RECORD,PM->from);
        } else FigureOutKey(RECORD);
        return IME_RT_BLOCKOUT;
      }
 }
  return IME_RT_REJECT;
}

int imeListArea(iomeRecord *RECORD,int x,int y)
{
  return 0;
}

int imeKeyArea(iomeRecord *RECORD,int x,int y)
{
  return 0;
}

int imeFocus(iomeRecord *RECORD)
{
  return 0;
}

int imeQuit(iomeRecord *RECORD)
{
  PhoneticMethod *PM = (PhoneticMethod*)RECORD->Owner;
  int n;
  for (n=0;n<41;n++) free(PM->CODE[n]);
  free(PM);
  free(RECORD->ListArea);
  free(RECORD->KeyArea);
  free(RECORD->Value);
  return 1;  
}

